import scipy.optimize as opt
import gc
import logging
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import pdb
import time
import sys
import scipy.linalg as lng

sys.path.append('..')

from MRA.MRATree import MRATree
import MRA.MRATools as mt




if __name__=='__main__':

    logging.basicConfig(format='%(asctime)s %(message)s', datefmt='%H:%M:%S',level=logging.INFO)

    np.random.seed(12)

    
    frac_obs = 0.3

    dim_x = 9
    dim_y = 1
    M=1; J=2; r0=1
    critDepth = 5

    

    

    ### simulate data ###

    sig = 1.0
    me_scale=1e-6
    kappa = 1.0


    
    if dim_y==1:
        locs = mt.genLocations(dim_x)
    else:
        locs = mt.genLocations2d( Nx=dim_x, Ny=dim_y )

    Sig = sig*mt.ExpCovFun(locs, l=kappa)
    SigC = np.matrix(lng.cholesky(Sig))

    x_raw = np.matrix(np.random.normal(size=(locs.shape[0],1)))
    x = SigC.T * x_raw


    R = me_scale
    Rc = np.sqrt(R) if isinstance(me_scale, float) else np.linalg.cholesky(R)
    eps = Rc * np.matrix(np.random.normal(size=(locs.shape[0],1)))
    y = x + eps

    # introducing missing data
    if np.ndim(y)==1:
        square = np.where(np.logical_and(np.logical_and(locs[:,0]>0.2, locs[:,0]<0.6), np.logical_and(locs[:,1]>0.2, locs[:,1]<0.6)))[0]
        not_obs_inds = np.random.choice(square, size=int(0.9*len(square)), replace=False)
        obs_inds = np.setdiff1d(np.arange(len(locs)), not_obs_inds)
    else:
        obs_inds = np.random.choice(dim_x*dim_y, int(dim_x*dim_y*frac_obs), replace=False)
    obs_inds=np.sort(obs_inds)
    y_obs = np.empty(np.shape(y)); y_obs[:] = np.NAN; y_obs[obs_inds] = y[obs_inds]



       
    ### MRA ###

    cov = lambda _locs1, _locs2: mt.ExpCovFun(_locs1, _locs2, l=kappa)
    mraTree = MRATree(locs, M, J, r0, critDepth, cov, y_obs, R)
    BP = mraTree.getBasisFunctionsMatrix(distr="posterior")
    
    


    ### kriging ###

    Sig = mt.ExpCovFun(locs, l=kappa)       
    H = np.matrix(np.eye(len(locs)))
    obsBool = np.isfinite(y_obs).ravel()
    H = H[obsBool,:]
    varY = H * Sig * H.T + R
    SigP = np.linalg.inv(np.linalg.inv(Sig) + H.T * (1/R)*np.eye(len(obs_inds)) * H)


    order = mraTree.root.getOrderFromLeaves()

        
    BP = mraTree.getBasisFunctionsMatrix(distr="posterior")
    mt.dispMat(BP, title="B")
    mt.dispMat(BP*BP.T, title="BB'")
    mt.dispMat(SigP[np.ix_(order, order)], title="Sigma")
    mt.dispMat(BP*BP.T - SigP[np.ix_(order, order)])



from pyMRA.MRANode import Node
from functools import reduce
import scipy.linalg
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import AxesGrid
import pdb

from pyMRA import MRATools as mt


class MRATree(object):


    def __init__(self, locs, M, J, r, critDepth, cov, obs, R):

        if J>1:
            if len(locs) < r*(1-J**(M))/(1-J):
                raise ValueError("Not enough grid points for the M, J, r you specified.")
        else:
            if len(locs) < r*(M+1):
                raise ValueError("Not enough grid points for the M, J, r you specified.")
            
        self.M = M
        self.J = J
        self.r = r
        self.locs = locs
        self.d = np.shape(locs)[1]
        obsM = np.matrix(obs) # make sure observations are a matrix; otherwise many operations will not work
        self.root = Node(None, 'r', locs, locs, M, J, r, critDepth, cov, obsM, R)
        self.obs_inds = np.where(np.logical_not(np.isnan(obs)))[0]





    def getLikelihood(self):

        return self.root.d + self.root.u


    
    

    def predict(self):

        xP = self.root.mean
        sdP = np.sqrt(self.root.var)
        return xP, sdP


        

        
        
    def getNodesBFS(self, groupByResolution=False):

        stack = [self.root]
        nodes = []
        while stack:
            n = stack.pop()
            
            if groupByResolution:
                if not nodes or nodes[-1][-1].res<n.res:
                    nodes.append([n])
                else:
                    nodes[-1].append(n)
            else:
                nodes.append(n)
                
            stack = n.children[::-1] + stack

        return nodes

    

    
    def getNodesDFS(self):
        
        stack = [self.root]
        nodes = []
        while stack:
            n = stack.pop()
            nodes.append(n)
            stack.extend(n.children[::-1])
            
        return nodes
    


    
    def drawKnots(self):

        fig = plt.figure()#figsize=plt.figaspect(0.2))
        nodes = self.getNodesBFS(groupByResolution=True)
        colors = ['#a6cee3','#b2df8a','#fb9a99','#ff7f00','#6a3d9a','#b15928']

        
        for m in range(self.M+1):

            if self.d==2:
                ax = fig.add_subplot(int(0.5*(self.M))+1, 2, m+1)
            else:
                ax = fig.add_subplot(self.M+1, 1, m+1)
                ax.set_ylim(-0.1, 2)

                
            # plot grid locations not used as knot before
            grid = [node.getGrid() for node in nodes[m]]

            for idx, region in enumerate(grid):
                col = colors[((idx+m) % len(colors))]
                if self.d==2:
                    ax.plot(region[:,0], region[:,1], marker='s', color=col, markersize='6', linestyle='None', label='grid')
                else:
                    ax.plot(region, np.zeros(len(region)), marker='s', color=col, markersize='6', linestyle='None', label='grid')
                        
            # plot knots
            knots = reduce(lambda a,b: np.vstack((a,b)), [node.locs[node.kInds] for node in nodes[m]])
            
            if self.d==2:
                ax.plot(knots[:,0], knots[:,1], marker='s', color='red', markersize='6', linestyle='None', label='knots')
            else:
                knots = ax.plot(knots[:,0], np.ones(len(knots)), marker='s', color='red', markersize='6', linestyle='None', label='knots')
                ax.get_yaxis().set_visible(False)
                ax.get_xaxis().set_visible(False)
            ax.set_title("resolution: %d" % m)

        plt.tight_layout()
        plt.show()      



    def drawBMatrix(self, distr="prior"):

        """
        Daws the B matrix. Its rows correspond to grid points and columns to
        hierarchically arranged basis functions
        """
        B = self.getBasisFunctionsMatrix(distr=distr)
        plt.matshow(B, cmap='Spectral')
        plt.colorbar()
        plt.tick_params(axis='both', which='both', bottom='off', top='off')
        plt.xlabel("%s basis functions" % distr, fontsize='x-large')
        plt.ylabel("grid points", fontsize='x-large')
        plt.show()





    def drawSparsityPat(self, distr="prior"):
        """
        draws the B matrix with ones whererver the corresponding element in the
        B matrix is nonzero and zeros otherwise.
        """
        if distr=="prior":
            B = self.getBasisFunctionsMatrix(distr="prior")
        elif distr=="posterior":
            B = self.getBasisFunctionsMatrix(distr="posterior")
        ind = np.where(np.abs(B)>1e-10)
        B1 = B; B1[ind] = 1
        Brr = B1
        #mt.dispMat(B1, cmap='binary', title="%s sparsity pattern" % distr, colorbar=False)
        mt.dispMat(Brr, cmap='binary', title="%s sparsity pattern" % distr, colorbar=False)
        #ind = np.where(np.abs(B.T*B)>1e-10)
        #BB1 = B.T * B; BB1[ind] = 1
        #Brr = BB1[:,::-1][::-1,:]
        #mt.dispMat(Brr, cmap='binary', colorbar=False)




    def drawBasisFunctions(self, distr="prior", order="root"):
        """
        plots basis functions
        """
        cmaps = [plt.cm.Blues, plt.cm.copper, plt.cm.Blues, plt.cm.copper]#plt.cm.Greys, plt.cm.YlOrBr, plt.cm.Purples, plt.cm.RdPu]

        ### 1D version
        if np.shape(self.locs)[1]==1 and self.M>0:
           
            B = self.getBasisFunctionsMatrix(groupByResolution=True, order=order, distr=distr)
            nodes = self.getNodesBFS(groupByResolution=True)
            fig = plt.figure(figsize=(8,6))
            
            for m in range(self.M+1):

                Bm = B[m]
                cmap = cmaps[m];
                ncol = Bm.shape[1]; offset=0.3*ncol
                
                ax = fig.add_subplot(self.M+1, 1, m+1)
                ax.set_ylim(top=1.1)
                ax.set_xlim(np.min(self.locs), np.max(self.locs))
                
                # draw partition lines
                for node_idx, node in enumerate(nodes[m]):
                    if np.min(node.locs)>np.min(self.locs):
                        ax.axvline(x=node.locs[0], linestyle='dashed', color='k', linewidth=1)

                # draw functions
                for col in range(Bm.shape[1]):
                    color = cmap(((offset+col)/(ncol+offset)))
                    ax.plot(self.locs, Bm[:,col], color=color, linewidth=2)
                    
                ax.tick_params(labelsize='large')
                ax.plot(self.locs, np.zeros(len(self.locs)), color='black', linewidth=2)
                
                if m<(len(B)-2):
                    ax.get_xaxis().set_visible(False)
                ax.set_title("resolution: %d" % m, fontsize='x-large')

            
            plt.tight_layout()
            plt.show()

        ### 2D version
        if np.shape(self.locs)[1]==2 and self.M:

            dim_x = len(np.unique(self.locs[:,0]))
            dim_y = len(np.unique(self.locs[:,1]))

            B = self.getBasisFunctionsMatrix(groupByResolution=True)

            for m, Bm in enumerate(B):

                if Bm.shape[1]>36:
                    continue
                nrows, ncols = mt.get_layout(m, self.J, self.r)
                
                fig = plt.figure()
                grid = AxesGrid(fig, 111,
                                nrows_ncols=(nrows, ncols),
                                axes_pad=0.1,
                                share_all=True,
                                label_mode="L",
                                cbar_location="right",
                                cbar_mode="single")

                for func, ax in zip(Bm.T, grid):
                    im = ax.imshow(np.array(func).reshape((dim_x, dim_y)), vmax=1, vmin=-0.1, cmap="coolwarm")
                    ax.get_xaxis().set_visible(False)
                    ax.get_yaxis().set_visible(False)

                plt.suptitle("resolution: %d" % m, fontsize="x-large")

                grid.cbar_axes[0].colorbar(im)

                plt.show()
                    


        

    def drawGridAndObs(self):

        """
        plot grid points and observation locations
        """
               
        if np.shape(self.locs)[1]==1:
            fig = plt.figure()
            ax = fig.add_subplot(111)

            ax.plot(self.locs, np.zeros(len(self.locs)), marker='o', color='black', markersize='3', linestyle='None', label="grid locations")
            ax.plot(self.locs[self.obs_inds], np.ones(len(self.obs_inds))*0.1, marker='o', color='red', markersize='4', linestyle='None', label="observations")
            ax.set_ylim([-.01, 0.2])
            ax.get_yaxis().set_visible(False)
            ax.legend(fontsize='large')
            plt.xticks(fontsize='large')
            plt.show()


            
        elif np.shape(self.locs)[1]==2 and self.M:

            if np.shape(self.locs)[1]>1:
                plt.scatter(self.locs[self.obs_inds,0], self.locs[self.obs_inds,1]);
            else:
                plt.scatter(self.locs[self.obs_inds,0], np.zeros(len(self.obs_inds)));
            plt.title("observation locations")
            plt.show()


            

        
    
    def getB_lk(self, callerID, k, l=None):
        """
        Queries node j_1, ... j_k for its covariance v_k with grid points in
        subregion j_1, ..., j_k, ..., j_l.

        Parameters
        ----------
        callerID : string
          ID of the node that asked for the information
        k : int
          resolution of the node on the caller path that contains the covariance
        l : int
          resolution of the node on the caller path with which the covariance
          should be calculated; l has to be between k and len(callerID)

        Returns
        -------
        Node:
          an instance of the Node class
        """

        kNode = self.getKNode(callerID, k)
        
        path = list(callerID[(k+1):] if not l else callerID[(k+1):(l+1)])[::-1]
        node = kNode
        inds = [np.arange(kNode.N)]
            
        while path:
            chId = node.children[int(path.pop())-1].ID
            inds.append( node.inds[chId] )
            node = node.children[int(chId[-1])-1]

        curInds = inds.pop()

        while inds:
            curInds = inds.pop()[curInds]
            
        return kNode.B[curInds,:]
    
        


    def getKNode(self, callerID, k):
        """
        Retrieves the k-th node on the path to the caller. For examples
        if caller ID is (j_1, j_2, j_3, j_4, j_5) and k=3 then the function
        returns (j_1, j_2, j_3). If k=0 then the root is returned. 
        This function should only be used by other nodes on the tree, not called
        from the outside.
        
        Parameters:
        -----------
        callerID : str
          ID of the node requesting the information
        k : int
          number of levels of the hierarchy down from the root

        Returns:
        --------
        node : Node
          the node with the requested ID
        """
        
        path = list(callerID[::-1])[:-1]
        node = self.root
        tempK = k
        
        while tempK>0:
            node = node.children[int(path.pop())-1]
            tempK = tempK-1
            
        return node




    
    def setPrior(self, xF, Sigma):

        self.root.calculatePrior(Sigma)




      

    def getBasisFunctionsMatrix(self, distr="prior", groupByResolution=False, order='root', timesKC=False):

        nodes = self.getNodesBFS(groupByResolution=True)
        r = self.root
        
        if distr=="prior":
            B = r.B
            if timesKC:
                B = B * r.kC
        elif distr=="posterior":
            B = r.BTil[self.root.res]
            if timesKC:
                B = B * r.kTilC

        if order=="leaves":
            leafOrder = r.getOrderFromLeaves()
            B = B[leafOrder,:]
            
        if groupByResolution:
            B = [B]

            
        for m_nodes in nodes[1:]:

            # figure out the order in which the grid points in children will be sorted
            if order=='leaves':
                orders=[]
                for n in m_nodes:
                    orders.append(n.getOrderFromLeaves())
            if order=='root':
                smallest_locs = np.array([np.min(n.locs[:,0]) for n in m_nodes])
                orders = np.argsort(smallest_locs)

                    
            Jr = self.r*len(m_nodes)
            if distr=="prior":
                if timesKC:
                    block_list = [np.array(n.B*n.kC) for n in m_nodes]
                else:
                    block_list = [np.array(n.B) for n in m_nodes]

            elif distr=="posterior":
                if timesKC:
                    block_list = [np.array(n.BTil[n.res] * n.kTilC) for n in m_nodes]
                else:
                    block_list = [np.array(n.BTil[n.res]) for n in m_nodes]


            # reorder blocks
            if order=="leaves":
                for node_num in range(len(m_nodes)):
                    block_list[node_num] = block_list[node_num][orders[node_num],:]
            if order=='root':
                block_list = [block_list[i] for i in orders]

                    
                
            Bm = scipy.linalg.block_diag(*block_list)
            if groupByResolution:
                B.append(Bm)
            else:
                B = np.hstack((B, Bm))
                
        return B



import scipy.optimize as opt
import gc
import logging
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import pdb
import time
import sys
import scipy.linalg as lng

sys.path.append('../..')

from pyMRA.MRATree import MRATree
from pyMRA import MRATools as mt




if __name__=='__main__':

    logging.basicConfig(format='%(asctime)s %(message)s', datefmt='%H:%M:%S',level=logging.INFO)

    np.random.seed(11)

    test_small = True
    diagnose = False
    krig = False
    compare=False
    find_params=False
    
    frac_obs = 0.4
    if test_small:
        dim_x = 10
        dim_y = 10
        M=1; J=4; r0=25
        critDepth = M+1
    else:
        dim_x = 100
        dim_y = 100
        M=4; J=2; r0=25
        critDepth = 6
    

    

    ### simulate data ###

    sig = 1.0
    me_scale=1e-1
    kappa = 0.3


    
    logging.info('=== simulating data ===')
    logging.info('simulating locations')
    if dim_y==1:
        locs = mt.genLocations(dim_x)
    else:
        locs = mt.genLocations2d( Nx=dim_x, Ny=dim_y )

    logging.info('calculating covariance matrix')
    if test_small:
        Sig = sig*mt.ExpCovFun(locs, l=kappa)
        #Sig = sig*mt.Matern32(locs, l=kappa, sig=sig)
        SigC = np.matrix(lng.cholesky(Sig))
    else:
        Sig = sig*mt.ExpCovFun(locs, l=kappa)
        SigC = np.matrix(lng.cholesky(Sig))
        np.save("SigC.npy", SigC)
        #SigC = np.load("SigC.npy")

    x_raw = np.matrix(np.random.normal(size=(locs.shape[0],1)))
    x = SigC.T * x_raw
    del SigC


    
    logging.info("simulating observations")
    R = me_scale
    Rc = np.sqrt(R) if isinstance(me_scale, float) else np.linalg.cholesky(R)
    eps = Rc * np.matrix(np.random.normal(size=(locs.shape[0],1)))
    y = x + eps

    # introducing missing data
    if np.ndim(y)==1:
        square = np.where(np.logical_and(np.logical_and(locs[:,0]>0.2, locs[:,0]<0.6), np.logical_and(locs[:,1]>0.2, locs[:,1]<0.6)))[0]
        not_obs_inds = np.random.choice(square, size=int(0.9*len(square)), replace=False)
        obs_inds = np.setdiff1d(np.arange(len(locs)), not_obs_inds)
    else:
        obs_inds = np.random.choice(dim_x*dim_y, int(dim_x*dim_y*frac_obs), replace=False)
    obs_inds=np.sort(obs_inds)
    y_obs = np.empty(np.shape(y)); y_obs[:] = np.NAN; y_obs[obs_inds] = y[obs_inds]
    del Rc
    gc.collect()





    
    
    
    
    ### MRA ###

    logging.info('=== starting MRA ===')
    start = time.time()
    cov = lambda _locs1, _locs2: mt.ExpCovFun(_locs1, _locs2, l=kappa)
    #cov = lambda _locs1, _locs2: mt.Matern32(_locs1, _locs2, l=kappa)
    mraTree = MRATree(locs, M, J, r0, critDepth, cov, y_obs, R)

    xP, sdP = mraTree.predict()
    sdP = sdP.reshape((dim_x, dim_y), order='A')
    #sdP = np.flipud(sdP.reshape((dim_x, dim_y), order='A'))
    MRATime = time.time()-start

    B = mraTree.getBasisFunctionsMatrix(distr="prior")
    
    logging.info('MRA predictions finished. It took {:.2}s'.format(MRATime))
    








    ### kriging ###

    if krig:
        logging.info('=== Starting ordinary kriging ===')

        start = time.time()
        Sig = mt.ExpCovFun(locs, l=kappa)
        
        H = np.matrix(np.eye(len(locs)))
        obsBool = np.isfinite(y_obs).ravel()
        H = H[obsBool,:]

        #if np.ndim(R)==1:
        #    R = np.diag(R)
        #elif isinstance(R, float):
        #    R = np.eye(len(obs_inds))*R

        varY = H * Sig * H.T + R

        SigP = np.linalg.inv(np.linalg.inv(Sig) + H.T * (1/R)*np.eye(len(obs_inds)) * H)
        sd = np.sqrt(np.diag(SigP))

        y_obs_k = np.zeros(np.shape(y_obs))
        y_obs_k[np.where(np.isnan(y_obs))] = 0

        xk = Sig*H.T*np.linalg.inv(varY)*H*y_obs_k
        sdk = sd.reshape((dim_x, dim_y), order='A')
        #sdk = np.flipud(sd.reshape((dim_x, dim_y), order='A'))
        regTime = time.time() - start

        logging.info('Kriging finished. It took {:.2}s'.format(regTime))
  
        #illustrate what simple kriging is about
        fig = plt.figure(figsize=plt.figaspect(0.2))
        axMean = fig.add_subplot(111)
        lineO = axMean.plot(locs[obs_inds], y[obs_inds], color='#deaa87', marker='o', linestyle='None', markersize='7', label='observations (Y)')
        lineP = axMean.plot(locs, x, color='#0000ff', linewidth=2, label='true state (X)')

        grid_ypos = np.min(np.vstack((np.min(x), y[obs_inds]))) - 0.1
        axMean.plot(locs, np.ones(len(locs))*grid_ypos, marker='o', color='#000000', markersize='7', linestyle='None', label='grid')
        
        axMean.legend(fontsize='x-large')
        plt.xticks(fontsize='x-large')
        plt.yticks(fontsize='x-large')
        plt.show()





   
    
    ### diagnostic plots ###

    if diagnose:
        mraTree.drawBMatrix("prior")
        mraTree.drawSparsityPat("prior")
        mraTree.drawBMatrix("posterior")
        mraTree.drawSparsityPat("posterior")
        
        #mraTree.drawBasisFunctions("prior")
        #mraTree.drawBasisFunctions("posterior")
        #mraTree.drawGridAndObs()
        mraTree.drawKnots()



        


    ### parameter optimization ###

    def likelihood(kappa):
    #def likelihood(params):
        """
        Evaluates the likelihood of the model for a given set of parameters
        params=(kappa, sigma, R)
        """
        #par = dict(zip(('kappa', 'sigma', 'R'), np.abs(params)))      
        #logging.debug("kappa=%f, sig=%f, R=%f" % (par['kappa'], par['sigma'], par['R']))

        cov = lambda _locs1, _locs2: mt.ExpCovFun(_locs1, _locs2, l=kappa)
        #cov = lambda _locs1, _locs2: mt.Matern32(_locs1, _locs2, l=par['kappa'], sig=1.0)
        mraTree = MRATree(locs, M, J, r0, critDepth, cov, y_obs, me_scale)
        #mraTree = MRATree(locs, M, J, r0, critDepth, cov, y_obs, np.abs(par['R']))
        lik = mraTree.getLikelihood()
        return( lik )

    if find_params:
        logging.info("=== finding optimal parameter values ===")
        start = time.time()
        xmin = opt.minimize(likelihood, [kappa], method='nelder-mead', options={'xtol':1e-1, 'disp':False})
        #xmin = opt.minimize(likelihood, [kappa, sig, me_scale], method='nelder-mead', options={'xtol':1e-2, 'disp':False})
        opttime = time.time() - start

        
        
        logging.info("ML estimates found. It took %ds" % opttime)
        logging.info(str(xmin))
        #logging.info(dict(zip(('kappa'),xmin['x'])))
        #logging.info(dict(zip(('kappa', 'sigma', 'R'),xmin['x'])))
    


        
    
    ### compare results ###

    if krig and compare:
        if dim_y>1:


            my_map = mpl.cm.get_cmap('Spectral')
            my_map.set_bad(color='grey')

            vmin = np.min(np.vstack((y_obs_k, x)))
            vmax = np.max(np.vstack((y_obs_k, x)))
            
            mt.dispMat(xP.reshape((dim_x, dim_y)), cmap='RdYlBu', title="predicted x")          
            mt.dispMat(x.reshape((dim_x, dim_y)), cmap='Spectral', title="true state (X)", vmax=vmax, vmin=vmin, fontsize='xx-large')
            if test_small:
                mt.dispMat(xk.reshape((dim_x, dim_y), order='A'), cmap='RdYlBu', title='kriged x')
                mt.dispMat((xP-xk).reshape((dim_x, dim_y)), cmap='coolwarm', title="predicted x - true x")
            mt.dispMat(y_obs.reshape((dim_x, dim_y), order='A'), cmap=my_map, title="observations (Y)", vmax=vmax, vmin=vmin, fontsize='xx-large')
            mt.dispMat(sdP, cmap='Reds', title="predicted sd")

        else:
            # compare MRA and kriging results
            fig = plt.figure(figsize=plt.figaspect(0.5))
            axMean = fig.add_subplot(121)
            lineP = axMean.plot(locs, xP, 'r-')
            lineO = axMean.plot(locs[obs_inds], y[obs_inds], 'ko', markersize='4')
            lineT = axMean.plot(locs, xk, linestyle='dashed', color="gray")
            axMean.set_title("mean")
               
            axSd = fig.add_subplot(122)
            #axSd.plot(locs, sdP, 'r-')
            #axSd.plot(locs, sd, linestyle='dashed', color="gray")
            axSd.plot(locs, sd-sdP.T.ravel(), 'r-')
            axSd.set_title("sd")
            plt.show()

            pdb.set_trace()



import logging
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import cm
import pdb
import time
import sys
import scipy.linalg as lng
#import scipy.sparse as sp

sys.path.append('../..')

from pyMRA.MRATree import MRATree
import pyMRA.MRATools as mt
#import models.diffusion


"""
This script is used to make predictions on a data set supplied in a csv
file with each row corresponding to one observation. The first two columns
represent the x and y coordinate, the third is the measurement and the
fourth is the truth - if available
"""



if __name__=='__main__':

    np.random.seed(11)

    #filename = '/home/marcin/MRF/data/Exp_Theta0.1_X100_Y100.csv'
    #filename = '/home/marcin/MRF/data/Exp_Theta0.1_X100_Y100_missing_all.csv'
    #filename = '/home/marcin/MRF/data/sat_temps.csv'
    filename = '/home/marcin/temp/test'
    logging.basicConfig(format='%(asctime)s %(message)s', datefmt='%H:%M:%S',level=logging.INFO)
    
    M=2; J=4; r0=2
    me_scale=1e-4
    critDepth = M+1

    
    with open(filename, 'r') as ipFile:
        csv = [line.strip().split(',') for line in ipFile.readlines()][1:]

    N = len(csv)

    locs=np.zeros((N, 2))
    y_obs = np.zeros((N, 1))
    y = np.zeros((N, 1))

    
    for idx, line in enumerate(csv):

        locs[idx, 1] = float(line[0])
        locs[idx, 0] = float(line[1])
        y_obs[idx, 0] = float(line[2]) if line[2]!='NA' else np.NAN
        if len(line)>3:
            y[idx, 0] = float(line[3])

    locs[:,0] = locs[:,0] - np.min(locs[:,0])
    locs[:,1] = locs[:,1] - np.min(locs[:,1])

    locs[:,0] = locs[:,0]/np.max(locs[:,0])
    locs[:,1] = locs[:,1]/np.max(locs[:,1])

    Nx = len(np.unique(locs[:,0])); Ny = len(np.unique(locs[:,1]))
    
    obs_mean = np.nanmean(y_obs)
    y_obs = y_obs - obs_mean
    y = y - obs_mean
    
    obs_inds = np.isfinite(y_obs).flatten()
    R = me_scale

    y_disp = y_obs.reshape((Nx, Ny))
    mt.dispMat(y_disp, cmap="Spectral", title="observations")

    sys.exit()


    
    logging.info("MRA started")
    start = time.time()
    cov = lambda _locs1, _locs2: mt.ExpCovFun(_locs1, _locs2, l=0.3)


    start = time.time()
    mraTree = MRATree(locs, M, J, r0, critDepth, cov, y_obs, R)
       
    yP, sdP = mraTree.predict()
    sdP = sdP.reshape((Nx, Ny))
    #sdP = np.flipud(sdP.reshape((Nx, Ny)))
    yP = yP.reshape((Nx, Ny))

    mraTime = time.time()-start
    logging.info( "MRA finished in {:.2}s".format(mraTime) )
    
    np.save("y-pred.npy", yP)


   
    
    ### compare results


    mt.dispMat(yP, cmap="Spectral", title="prediction")
    mt.dispMat(sdP, cmap="coolwarm", title="standard deviation", vmin=0.05, vmax=0.36)
    if np.any(y):
        y = y.reshape((Nx, Ny), order='A')
        mt.dispMat(y, cmap="Spectral", title="truth")
        mt.dispMat(yP - y, cmap="coolwarm", title="difference")

        





import os
import traceback
import time
import numpy as np
import multiprocessing






class detCalculator(multiprocessing.Process):
  

    def __init__(self, ID, matSize, noChildren, levelsLeft, parent_pipe):

        multiprocessing.Process.__init__(self)
        self.ID = ID
        self.parent_pipe = parent_pipe
        self.mat = np.random.uniform(size=(matSize, matSize))
        self.children = []
        self.pipes = []

        print("created node " + self.ID + ":, pid=" + str(os.getpid()))

        
        if levelsLeft:
            for chNo in range(noChildren):
                chID = self.ID + str(chNo+1)
                parent_end, child_end = multiprocessing.Pipe()
                
                chP = multiprocessing.Process( target=detCalculator, args=(chID, matSize, noChildren, levelsLeft-1, child_end) )
                self.children.append( chP )
                self.pipes.append( parent_end )
                chP.start()
                




    def run(self):

        while True:
            print("Node " + self.ID + ": " + "running the loop")
            try:
                msg = self.parent_pipe.recv()
                print("Node " + self.ID + ": " + msg)
                for ch in self.children:
                    print("ch:" + ": pid=" + str(ch.pid))
                for ch_pipe in self.pipes:
                    ch_pipe.send(msg)
                    print("sending message")
            except:
                print("error")
                traceback.print_exc()
                break
        print("outside of the loop")
        return

                
            

    # def run(self):

    #     data = []
    #     while True:
    #         try:
    #             msg = self.parent_pipe.recv()
    #             if msg is None:
    #                 self.kill_children()
                    
    #             elif msg=="getDet":
    #                 data.append( (self.ID, np.linalg.det(self.mat)) )

    #                 Nremain = len(self.children)
                    
    #                 while Nremain:
    #                     if not self.q.empty():
    #                         q.get()
    #                         Nremain = Nremain - 1
    #                         data.extend(chData)

    #                 for ch_pipe in self.pipes:
    #                     ch_pipe.send(msg)

                    

                        
    #             self.parent_pipe.send( data )
    #         except EOFError:
    #             break
        







if __name__=='__main__':

    matSize = 100
    noChildren=2
    levels = 1
    
    conn1, conn2 = multiprocessing.Pipe()
    newCalc = detCalculator('r', matSize, noChildren, levels, conn2)

    newCalc.start()
    conn1.send("getDet")


    



"""
  Some utility functions useful for MPI parallel programming
"""
from __future__ import print_function

from mpi4py import MPI

#=============================================================================
# I/O Utilities

def pprint(str="", end="\n", comm=MPI.COMM_WORLD):
    """Print for MPI parallel programs: Only rank 0 prints *str*."""
    if comm.rank == 0:
        print(str+end, end=' ') 



import scipy.optimize as opt
import gc
import logging
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import pdb
import time
import sys
import scipy.linalg as lng

from pyMRA.MRATree import MRATree
import pyMRA.MRATools as mt


if __name__ == '__main__':

    logging.basicConfig(format='%(asctime)s %(message)s', datefmt='%H:%M:%S',
                        level=logging.INFO)
    logging.info("===Unit tests for MRA===")

    ##### #1. MRA in 1d exponential with 0 resolutions is the same as kriging #####
    dim_x = 2
    M=0; J=1; r0=dim_x
    frac_obs = 0.5

    sig = 1.0
    me_scale=1e-10
    kappa = 1.0

    locs = mt.genLocations(dim_x)
    Sig = sig*mt.ExpCovFun(locs, l=kappa)
    SigC = np.matrix(lng.cholesky(Sig))

    x_raw = np.matrix(np.random.normal(size=(locs.shape[0],1)))
    x = SigC.T * x_raw

    R = me_scale
    Rc = np.sqrt(R) if isinstance(me_scale, float) else np.linalg.cholesky(R)
    eps = Rc * np.matrix(np.random.normal(size=(locs.shape[0],1)))
    y = x + eps

    obs_inds = np.random.choice(dim_x, int(dim_x*frac_obs), replace=False)
    obs_inds=np.sort(obs_inds)
    y_obs = np.empty(np.shape(y)); y_obs[:] = np.NAN; y_obs[obs_inds] = y[obs_inds]
    
    ### MRA ###
    cov = lambda _locs1, _locs2: mt.ExpCovFun(_locs1, _locs2, l=kappa)
    mraTree = MRATree(locs, M, J, r0, M+1, cov, y_obs, R)
    xP, sdP = mraTree.predict()
    sdP = sdP.reshape((dim_x, 1), order='A')
    
    ### kriging ###
    H = np.matrix(np.eye(len(locs)))
    obsBool = np.isfinite(y_obs).ravel()
    H = H[obsBool,:]
    R = np.eye(len(obs_inds))*me_scale
        
    varY = H * (Sig + R) * H.T
    SigP = np.linalg.inv(np.linalg.inv(Sig) + H.T * np.linalg.inv(R) * H)
    sd = np.sqrt(np.diag(SigP))

    y_obs[np.where(np.isnan(y_obs))] = 0
    xk = Sig*H.T*np.linalg.inv(varY)*H*y_obs
    sdk = sd.reshape((dim_x, 1), order='A')

    order = np.floor(np.log10(np.mean(np.abs(x))))
    assert np.max(np.abs(xP - xk)) < (10**(order-3)), \
    			"test #1 (1d exponential, 0 resolutions) not passing"
    assert np.max(np.abs(sdP - sdk)) < 10**(order-3), \
    			"test #1 (1d exponential, 0 resolutions) not passing"

    ####################################################
    
    ### #2. MRA in 1d exponential with 1 resolution, 
    ### knots on the boundary the same as kriging
    
    dim_x = 3
    M=1; J=2; r0=1
    frac_obs = 0.5

    sig = 1.0; me_scale=1e-10;  kappa = 1.0

    locs = mt.genLocations(dim_x)
    Sig = sig*mt.ExpCovFun(locs, l=kappa)
    SigC = np.matrix(lng.cholesky(Sig))

    x_raw = np.matrix(np.random.normal(size=(locs.shape[0],1)))
    x = SigC.T * x_raw

    R = me_scale
    if isinstance(me_scale, float):
        Rc = np.sqrt(R)
    else:
        Rc = np.linalg.cholesky(R)
    eps = Rc * np.matrix(np.random.normal(size=(locs.shape[0],1)))
    y = x + eps

    obs_inds = np.random.choice(dim_x, int(dim_x*frac_obs), replace=False)
    obs_inds=np.sort(obs_inds)
    y_obs = np.empty(np.shape(y)); y_obs[:] = np.NAN; y_obs[obs_inds] = y[obs_inds]
    
    ### MRA ###
    cov = lambda _locs1, _locs2: mt.ExpCovFun(_locs1, _locs2, l=kappa)
    mraTree = MRATree(locs, M, J, r0, M+1, cov, y_obs, R)

    xP, sdP = mraTree.predict()
    sdP = sdP.reshape((dim_x, 1), order='A')
  

    ### kriging ###
    H = np.matrix(np.eye(len(locs)))
    obsBool = np.isfinite(y_obs).ravel()
    H = H[obsBool,:]

    R = np.eye(len(obs_inds))*me_scale
        
    varY = H * (Sig + R) * H.T
    SigP = np.linalg.inv(np.linalg.inv(Sig) + H.T * np.linalg.inv(R) * H)
    sd = np.sqrt(np.diag(SigP))

    y_obs[np.where(np.isnan(y_obs))] = 0
    xk = Sig*H.T*np.linalg.inv(varY)*H*y_obs
    sdk = sd.reshape((dim_x, 1), order='A')

    order = np.floor(np.log10(np.mean(np.abs(x))))
    assert np.max(np.abs(xP - xk)) < (10**(order-3)), \
				"test #2 (1d exponential, 1 resolution) not passing"
    assert np.max(np.abs(sdP - sdk)) < 10**(order-3), \
    				"test #2 (1d exponential, 1 resolution) not passing"

    ########################################
    
   ############# #3. MRA works in 2d ############

    ### We don't plot the results here because we use the Matern covariance
    ### functions. It does not give us a screening effect so the MRA results are
    ### not exact. This is only to show that the code works properly (doesn't crash)
    ### when we deal with 2d domain.
    
    np.random.seed(12)
    frac_obs = 0.7

    dim_x = 10; dim_y = 10
    M=2; J=3; r0=2
    critDepth = 6

    sig=1.0; me_scale=1e-6; kappa=0.5

    ### simulate data ###
    locs = mt.genLocations2d(Nx=dim_x, Ny=dim_y)

    Sig = sig*mt.Matern32(locs, l=kappa, sig=sig)
    SigC = np.matrix(lng.cholesky(Sig))

    x_raw = np.matrix(np.random.normal(size=(locs.shape[0],1)))
    x = SigC.T * x_raw
    del SigC

    R = me_scale
    Rc = np.sqrt(R) if isinstance(me_scale, float) else np.linalg.cholesky(R)
    eps = Rc * np.matrix(np.random.normal(size=(locs.shape[0],1)))
    y = x + eps

    # missing data
    obs_inds = np.random.choice(dim_x*dim_y, int(dim_x*dim_y*frac_obs), replace=False)
    obs_inds=np.sort(obs_inds)
    y_obs = np.empty(np.shape(y)); y_obs[:] = np.NAN; y_obs[obs_inds] = y[obs_inds]
    del Rc
       
    
    ### MRA ###
    cov = lambda _locs1, _locs2: mt.Matern32(_locs1, _locs2, l=kappa, sig=sig)
    mraTree = MRATree(locs, M, J, r0, critDepth, cov, y_obs, R)

    xP, sdP = mraTree.predict()
    sdP = sdP.reshape((dim_x, dim_y), order='A')
    

    ### kriging ###
    Sig = mt.Matern32(locs, l=kappa, sig=sig)
        
    H = np.matrix(np.eye(dim_x*dim_y))
    obsBool = np.isfinite(y_obs).ravel()
    H = H[obsBool,:]
    
    if np.ndim(R)==1:
        R = np.diag(R)
    elif isinstance(R, float):
        R = np.eye(len(obs_inds))*R
        
    varY = H * Sig * H.T + R
    SigP = np.linalg.inv(np.linalg.inv(Sig) + H.T * np.linalg.inv(R) * H)
    sd = np.sqrt(np.diag(SigP))

    y_obs[np.where(np.isnan(y_obs))] = 0
    xk = Sig*H.T*np.linalg.inv(varY)*H*y_obs
    sdk = sd.reshape((dim_x, dim_y), order='A')
            
    ########################################
    logging.info("All tests passed")


#!/usr/bin/env python

from __future__ import division
from __future__ import print_function
import numpy as np
from mpi4py import MPI
from parutils import pprint


comm = MPI.COMM_WORLD



pprint("-"*78)
pprint(" Running on %d cores" % comm.size)
pprint("-"*78)


my_N = 4
N = my_N * comm.size


if comm.rank == 0:
    A = np.arange(N, dtype=np.float64)
else:
    A = np.empty(N, dtype=np.float64)

my_A = np.empty(my_N, dtype=np.float64)

# Scatter data into my_A arrays
comm.Scatter( [A, MPI.DOUBLE], [my_A, MPI.DOUBLE] )

pprint("After Scatter:")
for r in range(comm.size):
    if comm.rank == r:
        print("[%d] %s" % (comm.rank, my_A))
    comm.Barrier()

# Everybody is multiplying by 2
my_A *= 2

# Allgather data into A again
comm.Allgather( [my_A, MPI.DOUBLE], [A, MPI.DOUBLE] )

pprint("After Allgather:")
for r in range(comm.size):
    if comm.rank == r:
        print("[%d] %s" % (comm.rank, A))
    comm.Barrier()



import multiprocessing
import sys
import re

class ProcessWorker(multiprocessing.Process):
    """
    This class runs as a separate process to execute worker's commands in parallel
    Once launched, it remains running, monitoring the task queue, until "None" is sent
    """

    def __init__(self, task_q, result_q):
        multiprocessing.Process.__init__(self)
        self.task_q = task_q
        self.result_q = result_q
        return

    def run(self):
        """
        Overloaded function provided by multiprocessing.Process.  Called upon start() signal
        """
        proc_name = self.name
        print( '%s: Launched' % (proc_name))
        while True:
            next_task_list = self.task_q.get()
            if next_task is None:
                # Poison pill means shutdown
                print('%s: Exiting' % (proc_name))
                self.task_q.task_done()
                break
            next_task = next_task_list[0]
            print( '%s: %s' % (proc_name, next_task))
            args = next_task_list[1]
            kwargs = next_task_list[2]
            answer = next_task(*args, **kwargs)
            self.task_q.task_done()
            self.result_q.put(answer)
        return
# End of ProcessWorker class

class Worker(object):
    """
    Launches a child process to run commands from derived classes in separate processes,
    which sit and listen for something to do
    This base class is called by each derived worker
    """
    def __init__(self, config, index=None):
        self.config = config
        self.index = index

        # Launce the ProcessWorker for anything that has an index value
        if self.index is not None:
            self.task_q = multiprocessing.JoinableQueue()
            self.result_q = multiprocessing.Queue()

            self.process_worker = ProcessWorker(self.task_q, self.result_q)
            self.process_worker.start()
            print( "Got here")
            # Process should be running and listening for functions to execute
        return

    def enqueue_process(target):  # No self, since it is a decorator
        """
        Used to place an command target from this class object into the task_q
        NOTE: Any function decorated with this must use fetch_results() to get the
        target task's result value
        """
        def wrapper(self, *args, **kwargs):
            self.task_q.put([target, args, kwargs]) # FAIL: target is a class instance method and can't be pickled!
        return wrapper

    def fetch_results(self):
        """
        After all processes have been spawned by multiple modules, this command
        is called on each one to retreive the results of the call.
        This blocks until the execution of the item in the queue is complete
        """
        self.task_q.join()                          # Wait for it to to finish
        return self.result_q.get()                  # Return the result

    @enqueue_process
    def run_long_command(self, command):
        print( "I am running number % as process "%number, self.name )

        # In here, I will launch a subprocess to run a  long-running system command
        # p = Popen(command), etc
        # p.wait(), etc
        return 

    def close(self):
        self.task_q.put(None)
        self.task_q.join()

if __name__ == '__main__':
    config = ["some value", "something else"]
    index = 7
    workers = []
    for i in range(5):
        worker = Worker(config, index)
        worker.run_long_command("ls /")
        workers.append(worker)
    for worker in workers:
        worker.fetch_results()

    # Do more work... (this would actually be done in a distributor in another class)

    for worker in workers:
        worker.close() 



import logging
import pdb
import multiprocessing as mp

from ipyparallel import Client



class Node(object):
    """
    The basic building block of the tree
    """
    
    
    """
    critical depth property tells you at which level in the tree you want to parallelize.
    For example if you set it to a value that's greater than the number of levels in the
    hierarchy, parallelization will not kick in (useful for debugging). The other extreme
    would be to set it to 0: then each branch of the tree starting at the root will be 
    processed separately. This might not always be desireable though: if root has, say, only
    2 children, there will be only 2 processes runnning in parallel. But if every node has 
    2 children and we set critDepth to, say, 2, then we will have 8 separate processes.
    """
    critDepth = 2
    
    def __init__(self, ID, noChildren, height):

        self.ID = ID
        self.children = []
        self.depth = len(ID)-1
        self.msg = "no message"
        if height:
            self.leaf=False
            for j in range(noChildren):
                child = Node(ID + str(j+1), noChildren, height-1)
                self.children.append( child )
        else:
            self.leaf=True

            


    def assignMsg(self, msg):
        """
        This function looks silly, but this is just a template code; In the actual
        application this function is pretty complicated
        """
        self.msg = msg

            
            
    def setMsg(self,msg, pipe=None):
        """
        This is the recursive function that sets the message for all nodes.
        Once we are at the critical depth we create as many separate processes
        as there are children, and we assign a pipe to each process. Then we send
        the child and with the "setMsg" method through the pipe for parallel
        processing. There is one caveat here though. When a child process is being
        created it becomes an exact copy of its parent (here is where the excessive
        memory usage happens). This means that when its done with processing the
        child and the subtree this child is the root of, it has to return this
        entire subtree and swap it with the existing subtree. This is not great
        in terms of efficiency but it works.
        """
        self.assignMsg(msg)
        pipes = []; procs=[]

        
        for ch in self.children:
            if self.depth==self.critDepth:
                parPipe, chPipe = mp.Pipe()
                chP = mp.Process(target=ch.setMsg, args=(msg,chPipe))
                chP.start()
                pipes.append(parPipe)
                procs.append(chP)
            else:            
                ch.setMsg(msg)


        if self.depth==self.critDepth:
            #if this is where parallelization is performed, receive the subtree
            #through the appropriate pipe and join the process.
            for idx, pipe in enumerate(pipes):
                self.children[idx] = pipe.recv()
                procs[idx].join()
                

        if self.depth==(self.critDepth+1):
            # here the subtree is returned - sent through the pipe
            pipe.send(self)

            
            

            
    def getMsg(self):
        """
        This functions returns the string that has messages from the entire
        subtree starting at this node
        """
        strOut = self.depth*'\t'+'-> %s: %s\n' % (self.ID, self.msg)
        strOut = strOut + '\n'.join([ch.getMsg() for ch in self.children])
        if self.leaf: # if it's a leaf we remove the extra newline character
            strOut = strOut[:-1]
        return strOut



    
if __name__=='__main__':
    
    MaxDepth = 2
    noChildren = 2

    root = Node('r', noChildren, MaxDepth)

    print("Before the message is set")
    print(root.getMsg())

    
    root.setMsg("I am groot")
    print("After setting the message:")
    print(root.getMsg())



#import weakref
import os
import pdb
import numpy as np
from scipy.spatial.distance import squareform, pdist, cdist
import matplotlib as mpl
import matplotlib.pyplot as plt
import scipy.stats as sts
import numpy.linalg as lng





def get_layout(m,J,r):

    total = r*(J**m)
    if total<=2:
        tup = (1,2)
    elif total<=3:
        tup = (1,3)
    elif total<=4:
        tup = (1,4)
    elif total<=6:
        tup = (2,3)
    elif total<=8:
        tup = (2,4)
    elif total<=9:
        tup = (3,3)
    elif total<=12:
        tup = (3,4)
    elif total<=15:
        tup = (3,5)
    elif total==16:
        tup = (4,4)
    elif total<=18:
        tup = (3,6)
    elif total<=20:
        tup = (4,5)
    elif total<=24:
        tup = (4,6)
    elif total==25:
        tup = (5,5)
    elif total==28:
        tup = (4,7)
    elif total<=30:
        tup = (5,6)
    elif total<=35:
        tup = (5,7)
    elif total==36:
        tup = (6,6)
    else:
        raise ValueError("Too many functions to plot")
    return tup




def MSE(xPred, Sigma, xTrue=0):

    N = xPred.shape[0]
    err = np.sqrt((lng.norm(xPred - xTrue)**2)/N)
    
    return err




def KLdiv(mu0, mu1, Sig0, Sig1):
    
    n = Sig0.shape[0]

    Sig0c = np.linalg.cholesky(Sig0)
    Sig1c = np.linalg.cholesky(Sig1)

    M = np.linalg.solve(Sig1c, Sig0c)
    traceterm = sum(np.diag(M.T*M)) - n
    
    logdetterm = 2*(np.log(np.linalg.det(Sig1c))) - 2*(np.log(np.linalg.det(Sig0c)))

    meandiff = mu1 - mu0
    temp = np.linalg.solve(Sig1c, meandiff)
    meanterm = np.linalg.norm(temp,2)**2

    kldiv = 0.5*( traceterm + logdetterm + meanterm)
    return kldiv
    



# Calculates the normal log score
def logscore(obs, muPred, SigPred):


    y = np.array(obs).ravel()
    obs_inds = np.logical_not(np.isnan(y))

    yObs = y[obs_inds]
    muObs = np.array(muPred).ravel()[obs_inds]
    SigObs = np.array(SigPred)[np.ix_(obs_inds, obs_inds)]

    # W, V = np.linalg.eigh(SigPred)
    
    # lW = np.log(W)

    # x = (V.T * (xPred-xTrue))
    # N = len(x)

    #return -N/2 * np.sum(np.log(W)) - x.T * np.diag(1/W) * x
    return sts.multivariate_normal.logpdf(yObs, mean=muObs, cov=SigObs)




# calculates a matrix with distances between points of v1 and v2
def dist(v1, v2=None):

    if not np.any(v2):
        v2=v1
        
    D = np.zeros([len(v1), len(v2)])
    for idx_k, k in enumerate(v1):
        for idx_l, l in enumerate(v2):
            D[idx_k, idx_l] = np.abs(k-l)
    return(D)



def filterNNZ(X):
    newX = X
    indNNZ = np.where(X!=0)
    newX[indNNZ] = 1
    return newX



# displays a matrix
def dispMat(mat, title="", cmap=None, fName=None, vmin=None, vmax=None, colorbar=True):

    fig = plt.matshow(mat, cmap=cmap, vmin=vmin, vmax=vmax)
    if colorbar:
        plt.colorbar()
    if title:
        plt.title(title)
    if fName:
        plt.savefig(fName, dpi=400)
    fig.axes.get_xaxis().set_visible(False)
    fig.axes.get_yaxis().set_visible(False)
        

    plt.show()




    
    
# generates locations on a given domain
def genLocations(NGrid, lb=0, ub=1, random=False):

    if random:
        locs_all = np.random.uniform(lb, ub, NGrid)
    else:
        locs_all = np.linspace(lb, ub, num=NGrid+1)[1:]

    return(locs_all.reshape((NGrid, 1)))




def genLocations2d(Nx, lbx=0, ubx=1, Ny=0, lby=0, uby=1):

    if not Ny:
        Ny=Nx
    
    xx, yy = np.meshgrid(np.linspace(lbx, ubx, num=Nx), np.linspace(lby, uby, num=Ny))
    X = np.hstack((xx.flatten().reshape(Nx*Ny, 1), yy.flatten().reshape(Nx*Ny,1)))

    return X



def genClusters(N, k):

    NperK = int(N/k)
    points = np.empty((0,2))
    
    for cluster in range(k):
        pts = np.random.normal(loc=np.random.uniform(size=2), scale=np.random.uniform(low=0.1, high=0.2), size=(NperK,2))
        points = np.vstack((points, pts))

    for r in range(N- k*NperK):
        pts = np.random.uniform(size=2)
        points = np.vstack((points, pts))

    return points



def ExpCovFunCirc(locs, l=1):

    locs = locs if np.ndim(locs)==2 else np.reshape(locs, [len(locs), 1])

    xv, yv = np.meshgrid(locs, locs)
    m = np.minimum(xv, yv)
    M = np.maximum(xv, yv)
    dist = np.minimum(M - m, m + 1-M)

    covMat = np.exp(-dist/l)
    return(covMat)

    



# a covariance function

def ExpCovFun(locs, locs2=np.array([]), l=1, circular=False):

    locs = locs if np.ndim(locs)==2 else np.reshape(locs, [len(locs), 1])

    if circular:
        xv, yv = np.meshgrid(locs, locs)
        m = np.minimum(xv, yv)
        M = np.maximum(xv, yv)
        dist = np.minimum(M - m, m + 1-M)
    else:
        if len(locs2):
            dist = np.matrix(cdist(locs, locs2))
        else:
            dist = np.matrix(squareform(pdist(locs)))
    covMat = np.exp(-dist/l)


    return(covMat)





def Matern32(locs, locs2=np.array([]), l=1, sig=1):

    if len(locs2):
        dist = cdist(locs, locs2)
    else:
        dist = squareform(pdist(locs))

    covMat = sig*(1 + np.sqrt(3)*dist/l) * np.exp(-np.sqrt(3)*dist/l)

    return(np.matrix(covMat))




# a tapering function
def KanterCovFun(locs, radius=1):

    locs = locs if np.ndim(locs)==2 else np.reshape(locs, [len(locs), 1])
    
    D = squareform(pdist(locs))/radius
    piD2 = 2*np.pi*D
    R = (1 - D)*np.sin(piD2)/piD2 + 1/np.pi * (1-np.cos(piD2))/piD2
    R[D>1]=0
    R[D==0]=1
    return R




def simulate1D(locs, CFun, mean=None, seed=None, domainIsCircular=False):
    """
    Simulate from a 1-D random Gaussian field

    Parameters
    ----------
    locs : numpy.array
      1-D array with locations at which to simulate the process
    CFun : function
      covariance function to be used for simulation
    mean : function
      mean function of the process
    seed : int
      seed for numpy.random

    Returns
    -------
    y : numpy.array
      values of the process at the given locations
    """
    
    NGrid = len(locs)
    if seed:
        np.random.seed(seed)
    trueCovMat = CFun(locs, circular=True)
    covChol = lng.cholesky(trueCovMat)

    if not mean:
        mean = np.zeros([NGrid,1])
    else:
        assert(np.shape(mean)[0]==NGrid and np.shape(mean)[1]==1)
    
    innov = np.reshape(np.random.normal(size=NGrid), [NGrid, 1]) + mean
    y = np.dot(covChol, innov)
    return(y)
    



import numpy as np
import sys
import pdb
import multiprocessing as mp
from mpi4py import MPI


class Node(object):

    
    def __init__(self, ID, NCh, levelsFromLeaves):

        self.ID = ID
        self.children = []
        self.depth = len(self.ID)-1

        dim = 300
        self.mat = np.random.normal(size=dim*dim).reshape((dim, dim))
        self.inv = np.linalg.inv(self.mat)

        
        if levelsFromLeaves:

            if self.depth==2:
                self.pool = mp.Pool(processes=NCh)
                self.children = self.pool.starmap(Node, [(self.ID + str(j+1), NCh, levelsFromLeaves-1) for j in range(NCh)])

            else:
                for j in range(NCh):
                    chID = self.ID + str(j+1)
                    ch = Node(chID, NCh, levelsFromLeaves-1)
                    self.children.append(ch)
        


        

def times(x, y):

    return x*y

                    



if __name__=='__main__':

    M = 4
    J = 2

    root = Node('r', J, M)

    pdb.set_trace()
    sys.exit()
    
    pool = mp.Pool(processes=4)
    A = list([(i, i+1) for i in range(10)])
    results = pool.starmap(times, A)
    
    print(results)



import numpy as np
import pdb
import sys

sys.path.append('..')

import MRFTools as mt
import matplotlib.pyplot as plt



"""
This script tests an implementation of the ExpCovFunCric
function. It displays the covariance matrix and lets the 
user examine its pattern
"""


if __name__=='__main__':

    locs = np.linspace(start=0, stop=1, num=11)[:-1]
    #locs = np.array([0, 0.25, 0.5, 0.75])
    
    xv, yv = np.meshgrid(locs, locs)
    m = np.minimum(xv, yv)
    M = np.maximum(xv, yv)
    dist = np.minimum(M - m, m + 1-M)
    dist2 = np.roll(dist, 3, axis=1)
   
    S = mt.ExpCovFunCirc(locs)
    mt.dispMat(S)
    



import gc
import numpy_indexed as npi
import multiprocessing as mp
import logging
import pdb
import scipy.linalg as lng
from numpy.linalg import slogdet
import numpy as np
from sklearn.cluster import KMeans
from scipy.spatial.distance import *
import scipy
from pyMRA import MRATools as mt


class Node(object):

    #@profile
    def __init__(self, parent, ID, locs, notKnots, levelsFromLeaves,
                 J, r, critDepth, cov, obs, R, pipe=None):

        self.parent = parent
        self.children = []
        self.ID = ID
        self.res = len(ID)-1
        self.N = len(locs)
        self.locs = locs
        self.inds = {}
        self.critDepth = critDepth
        self.leaf=(not bool(levelsFromLeaves))
        
        if not self.leaf and len(notKnots)>max(r,J):
            self.knots, self.kInds = self._getKnotsInds(r, notKnots, random=False)
        else: 
            self.knots = notKnots
            self.leaf=True
            self.kInds = np.arange(self.N)[np.flatnonzero(npi.contains(notKnots, self.locs))]          

        self.calculatePrior(cov)
        
        if not self.leaf and len(notKnots)>max(r,J):
            
            newNotKnots = notKnots[~npi.contains(self.knots, notKnots),:]
            # if there is fewer "spare" locations than splits, then make as many splits as there are locations
            #minJ = min(J, len(newNotKnots)) 
            splits = self._getJSplits(J)
            #splits = self._getSplits()
            #splits = self._getSplitsRobust(newNotKnots)


            
            if self.res==self.critDepth:
                pipes=[]; procs=[]

            for j in range(len(splits)):
                chID = self.ID + str(j+1)
                self.inds[chID] = splits[j]
                
                chLocs = locs[self.inds[chID]]

                if isinstance(cov, np.matrix):
                    C = cov[np.ix_(self.inds[chID], self.kInds)]
                    covCh = cov[np.ix_(self.inds[chID], self.inds[chID])] - C * self.k * C.T
                elif len(locs)<=1e4: #if the child covariance matrix will not be too large, calculate it directly
                    C = cov(chLocs, self.knots)
                    covCh = cov(chLocs, chLocs) - C * self.k * C.T
                else:
                    covCh = lambda _locs1, _locs2: cov(_locs1, _locs2) - cov(_locs1, self.knots) * self.k * cov(self.knots, _locs2)

                chObs = obs[self.inds[chID]]
                chNotKnots = chLocs[npi.contains(newNotKnots, chLocs)]
                
                if isinstance(R, float):
                    chR = R
                else:
                    chR = R[np.ix_(self.inds[chID], self.inds[chID])]
                
                if self.res==self.critDepth:
                    parPipe, chPipe = mp.Pipe()
                    chP = mp.Process(target=Node, args=(self, chID, chLocs, chNotKnots, levelsFromLeaves-1, J, r, critDepth, covCh, chObs, chR, chPipe))
                    chP.start()
                    pipes.append(parPipe)
                    procs.append(chP)
                else:
                    newChild = Node(self, chID, chLocs, chNotKnots, levelsFromLeaves-1, J, r, critDepth, covCh, chObs, chR)
                    self.children.append( newChild )
                   
                    
        if self.res==self.critDepth:
            for idx, pipe in enumerate(pipes):
                self.children.append( pipe.recv() )
                procs[idx].join()

        self.calculatePosterior(obs, R)
        
        #while self.children:
        #    ch = self.children.pop()
        #    del ch
        #gc.collect()

        
        if self.res==(self.critDepth+1):
            pipe.send(self)

        

    def getLeaves(self):

        leaves = []
        if self.leaf:
            leaves.append(self)
        else:
            for ch in self.children:
                leaves.extend( ch.getLeaves() )

        return leaves

            


    
    def getOrderFromLeaves(self):

        leaves = self.getLeaves()

        
        def getLeafInds(leafID):
            path = list(leafID[(self.res+1):])[::-1]
            node = self
            inds = [np.arange(self.N)]
            
            while path:
                chId = node.children[int(path.pop())-1].ID
                inds.append( node.inds[chId] )
                node = node.children[int(chId[-1])-1]

            curInds = inds.pop()

            while inds:
                curInds = inds.pop()[curInds]

            return curInds

        leavesInds = [(leaf.ID, getLeafInds(leaf.ID)) for leaf in leaves]
        order = np.concatenate([tup[1] for tup in leavesInds])
        
        return order


            


    def getGrid(self):
        grid = self.knots
        for ch in self.children:
            chGrid = ch.getGrid()
            grid = np.vstack((grid, chGrid))
        return grid




            


    #@profile
    def _getKnotsInds(self, r, notKnots, random=False):

        if np.shape(self.locs)[1]==1:
            locs1d = self.locs.ravel()
            #knots = [np.percentile(locs1d, 100.0*i/(r+1), interpolation='nearest') for i in range(r+2)][1:-1]
            #knots = [np.percentile(notKnots1d, 100.0*i/(r+1), interpolation='nearest') for i in range(r+2)][1:-1]
            notKnots1d = notKnots.ravel()
            knots = [np.percentile(notKnots1d, 100.0*i/(r+1), interpolation='nearest') for i in range(r+2)][1:-1]
            kInds = np.arange(self.N)[np.flatnonzero(npi.contains(knots, locs1d))]
            

        else:
            if random and self.res>=0:
                inds = np.random.choice(np.arange(len(notKnots)), size=r, replace=False)
                knots = notKnots[inds]
            else:
                kmeans = KMeans(n_clusters=r, random_state=0).fit(notKnots)
                C = kmeans.cluster_centers_
                D = cdist(notKnots, C)
                knots = np.zeros((r, self.locs.shape[1]))
                for centr in range(r):
                    ind = np.argmin(D[:,centr])
                    knots[centr,:] = notKnots[ind]

                    kInds = np.arange(self.N)[np.flatnonzero(npi.contains(knots, self.locs))]
        knots = self.locs[kInds] # this is to ensure that both kInds and knots are in the same order
        return knots, kInds
                                          

    
    
    

    #@profile
    def _getSplits(self):

        """
        If locations are 1D, splits into three segments with the same number of grid points.
        Otherwise splits into four segments, splitting in half along each axis.
        """

        if np.shape(self.locs)[1]==1:

            perc = np.percentile(self.locs, (33,66))
            
            locs_0 = np.where(self.locs[:,0]<perc[0])[0]
            locs_1 = np.where(np.logical_and(self.locs[:,0]>perc[0], self.locs[:,0]<perc[1]))[0]
            locs_2 = np.where(self.locs[:,0]>perc[1])[0]

            subdomains = [locs_0, locs_1, locs_2]

        else:
        
            means = np.mean(self.locs, axis=0)
        
            locs_00 = np.where(np.logical_and(self.locs[:,0]<=means[0], self.locs[:,1]<=means[1]))[0]
            locs_01 = np.where(np.logical_and(self.locs[:,0]<=means[0], self.locs[:,1]>means[1]))[0]
            locs_10 = np.where(np.logical_and(self.locs[:,0]>means[0], self.locs[:,1]<=means[1]))[0]
            locs_11 = np.where(np.logical_and(self.locs[:,0]>means[0], self.locs[:,1]>means[1]))[0]
            
            subdomains = [locs_00, locs_01, locs_10, locs_11]

        
        return subdomains






    def _getSplitsRobust(self, notKnots):

        """
        Works like _getSplits but accounts for the fact that there might be few grid points along
        a certain dimension. It first splits the points that are not the "middle" points (along x, y
        or both). Then it splits the "ties" into 4 parts. This is useful when we are dealing with, say
        a 3x3 grid and want to split it into four subregions. If we didn't use the robust method
        we might end up with one region that has 4 points, two regions with 2 points and one region with only
        a single point. 

        Useful only in 2D
        """
        med = np.median(notKnots, axis=0)

        all_ties = notKnots[np.where(np.logical_and(notKnots[:,0]==med[0], notKnots[:,1]==med[1]))]
        ties_inds = np.arange(self.N)[np.flatnonzero(npi.contains(all_ties, self.locs))]
        
        xy_ties = np.where(np.logical_and(notKnots[:,0]==med[0], notKnots[:,1]==med[1]))[0].ravel()
        x_ties = np.where(notKnots[:,0]==med[0])[0]
        y_ties = np.where(notKnots[:,1]==med[1])[0]
        
        pure_x_ties = np.setdiff1d(x_ties, xy_ties)
        pure_y_ties = np.setdiff1d(y_ties, xy_ties)
        array_xy_ties = np.array_split(xy_ties, 4)
                           
        locs_list = [np.array_split(xy_ties, 4), np.array_split(pure_x_ties, 4),  np.array_split(pure_y_ties, 4)[::-1]]
        locs = []
        for i in range(4):
            locs.append( np.hstack((locs_list[0][i], locs_list[1][i], locs_list[2][i])) )       
      
        locs[0] = np.setdiff1d(np.where(np.logical_and(self.locs[:,0] < med[0],  self.locs[:,1] < med[1]))[0], ties_inds)
        locs[1] = np.setdiff1d(np.where(np.logical_and(self.locs[:,0] < med[0],  self.locs[:,1] > med[1]))[0], ties_inds)
        locs[2] = np.setdiff1d(np.where(np.logical_and(self.locs[:,0] > med[0],  self.locs[:,1] < med[1]))[0], ties_inds)
        locs[3] = np.setdiff1d(np.where(np.logical_and(self.locs[:,0] > med[0],  self.locs[:,1] > med[1]))[0], ties_inds)       
            
        return locs


    

    


    def _getJSplits(self, J):

        # if:
        # ** J=r+1, or if the number of knots is one less that the number
        #    of partitions,
        # ** we are in 1d
        # ** there is enough grid points left
        # then we partition such that the knots are at the boundary
        # ---- write the code for this ----
        # else:
        #    do k-means etc

        r = len(self.kInds)
        cond1 = J==r+1
        cond2 = self.locs.shape[1]==1
        cond3 = False
        
        if cond1 and cond2 and cond3:
            subdomains = [1]

        else:
            subDomains = []
            kmeans = KMeans(n_clusters=J, random_state=0).fit(self.locs)
            all_labels=kmeans.labels_
            for j in range(J):
                inds = np.where(all_labels==j)[0]
                if len(inds):
                    subDomains.append(inds)

            if self.locs.shape[1]==1:
                subDomains = sorted(subDomains, key=lambda _arr: np.min(_arr))

        return subDomains





    def _getB_lk(self, l):
 
        node = self
        inds = np.arange(self.N)
        while node.res>l:
            chID = node.ID
            node = node.parent
            inds = node.inds[chID][inds]
 
        return node.B[inds,:]



    

    def _getRowOrder(self):

        
        if True:#not self.children:
            return np.arange(len(self.locs))
        
        allInds = np.array([])
        for ch in self.children:
            allInds = np.hstack((allInds, self.inds[ch.ID]))

        order = np.argsort(allInds)
        return order
    


        
    #@profile
    def calculatePrior(self, cov):
        #order = self._getRowOrder()
        logging.debug("Node %s: calculate prior" % self.ID)
        if isinstance(cov, np.matrix):
            self.B = cov[:,self.kInds]
        else:
            self.B = cov(self.locs, self.knots)
        self.kInv = self.B[self.kInds,:]
        try:
            self.k = np.linalg.inv(self.kInv)
        except:
            logging.critical("Problem with the knots!")
            pdb.set_trace()    
        self.kC = np.linalg.cholesky(self.k)

        #self.B = self.B[order,:]
        
        logging.debug("Node %s: finished calculating prior" % self.ID)






    #@profile
    def calculatePosterior(self, obs, R):
        logging.debug("Node %s: start calculating posterior" % self.ID)

        
        # calculate A and omega
        self.A = []
        omg = []

        if self.leaf:


            ################     proposed change to posterior calculations     ################
            obsInds = np.array(np.isfinite(obs)).ravel()
            
            H = np.matrix(np.eye(len(obs)))[obsInds,:]
            
            Rmat = np.matrix(R*np.eye(sum(obsInds)))
                
            HRinvH = (1/R)*H.T*H
            HRinvObs = H.T*(1/R)*obs[obsInds]

            B_lk = [ self._getB_lk(k) for k in range(self.res+1) ]
            
            for k in range(self.res+1):
                self.A.append( [] )
                omg.append( B_lk[k].T * HRinvObs )
                for l in range(self.res+1):
                    self.A[k].append( B_lk[k].T * HRinvH * B_lk[l] )
            
        else:           
                
            for k in range(self.res+1):    
                subList=[ch.omgTil[k] for ch in self.children]
                omg.append( sum(subList) )
                self.A.append([])
                for l in range(self.res+1):
                    subList = [ch.ATil[k][l] for ch in self.children]
                    self.A[k].append( sum(subList) )                    

   
        
        self.kTil = np.matrix(lng.inv(self.kInv + self.A[self.res][self.res]))
        self.kTilInv = lng.inv(self.kTil)

        
        
        #likelihood:
        if self.leaf:

            self.u = -omg[self.res].T * self.kTil * omg[self.res] + obs[obsInds].T * (1/R) * obs[obsInds]
            if np.max(np.shape(self.u))>1:
                pdb.set_trace()
            
            sgnTil, logdetTil = slogdet(self.kTilInv)
            sgn, logdet = slogdet(self.kInv)
            sgnR, logdetR = slogdet(Rmat)
            self.d = logdetTil - logdet + logdetR


        else:
            self.d = -np.log(np.linalg.det(self.kTil)) - np.log(np.linalg.det(self.kInv))
            self.u = - omg[self.res].T * self.kTil * omg[self.res]

            for ch in self.children:
                self.d += ch.d
                try:
                    self.u += ch.u
                except:
                    pdb.set_trace()


        

        # calculate A-tilde and omega-tilde
        self.ATil = []
        self.omgTil = []
        for k in range(self.res):
            self.ATil.append( [] )
            self.omgTil.append( omg[k] - self.A[k][self.res] * self.kTil * omg[self.res] )
            for l in range(self.res+1):
                self.ATil[k].append( self.A[k][l] - self.A[k][self.res] * self.kTil * self.A[self.res][l] )



                
        # calculate B-tilde
        order = self._getRowOrder()
        self.BTil = []
        if self.leaf:
             self.BTil = [self._getB_lk(k) for k in range(self.res+1)]
        else:
            for k in range(self.res+1):
                self.BTil.append( np.matrix(np.zeros((self.N, len(self.kInds)))) )
                for ch in self.children:
                    chInds = self.inds[ch.ID]
                    self.BTil[k][chInds,:] = ch.BTil[k] - ch.BTil[ch.res] * ch.kTil * ch.A[ch.res][k]



        


        
                    

                
        W, V = np.linalg.eigh(self.kTil)
        negInd = np.where(W<0)[0]
        W[negInd] = -W[negInd]
        self.kTilC = np.matrix(V) * np.matrix(np.diag(np.sqrt(W)))

        # calculate the moments
        self.mean = self.BTil[self.res] * self.kTil * omg[self.res]
        self.var = np.linalg.norm(self.BTil[self.res] * self.kTilC, axis=1)**2 # we calculate only the diagonal to save memory


        
      
        # add contribution from children
        for ch in self.children:
            chInds = self.inds[ch.ID]
            self.mean[chInds,:] += ch.mean
            self.var[chInds] += ch.var

            
        logging.debug("Node %s: finished calculating posterior" % self.ID)





        



from setuptools import setup
setup(
    name = 'pyMRA',
    packages = ['pyMRA', 'pyMRA.data'],
    version = '0.7.3',
    description = 'Multi-resolution approximation for spatial Gaussian processes',
    author = 'Marcin Jurek',
    author_email = 'marcinjurek1988@gmail.com',
    url = 'https://github.com/marcinjurek/pyMRA',
    include_package_data=True,
    install_requires = ['numpy', 'matplotlib', 'scipy', 'numpy_indexed', 'sklearn'],
    #download_url = 'https://github.com/marcinjurek/pyMRA/archive/0.1.tar.gz',
    keywords = ['Gaussian process', 'multi-resolution', 'sparse', 'approximation', 'spatial', 'statistics'],
    classifiers = [],
)



import pyMRA
import os
import numpy as np

def load_data(size="small", include_truth=False):

    if size not in ["small", "large"]:
        raise ValueError("size has to be 'small' or 'large'")

    packagedir = pyMRA.__path__[0]
    dirname = os.path.join(os.path.dirname(packagedir), 'pyMRA', 'data', size)
    y_obs = np.load(os.path.join(dirname, 'y_obs.npy'))
    locs = np.load(os.path.join(dirname, 'locs.npy'))

    if include_truth:
        y = np.load(os.path.join(dirname, 'y.npy'))
        return( y, locs, y_obs )

    return( y, locs )
